from __future__ import annotations

import asyncio
from concurrent.futures import ThreadPoolExecutor
from functools import partial
from typing import TYPE_CHECKING, Any

from adaptive_scheduler.utils import _now, _serialize_to_b64, sleep_unless_task_is_done

from .base_manager import BaseManager
from .common import MaxRestartsReachedError, log

if TYPE_CHECKING:
    from adaptive_scheduler.scheduler import BaseScheduler
    from adaptive_scheduler.utils import (
        _DATAFRAME_FORMATS,
        LOKY_START_METHODS,
        GoalTypes,
    )

    from .database_manager import DatabaseManager


def command_line_options(
    *,
    scheduler: BaseScheduler,
    database_manager: DatabaseManager,
    runner_kwargs: dict[str, Any] | None = None,
    goal: GoalTypes,
    log_interval: float = 60,
    save_interval: float = 300,
    save_dataframe: bool = True,
    dataframe_format: _DATAFRAME_FORMATS = "pickle",
    loky_start_method: LOKY_START_METHODS = "loky",
) -> dict[str, Any]:
    """Return the command line options for the job_script.

    Parameters
    ----------
    scheduler
        A scheduler instance from `adaptive_scheduler.scheduler`.
    database_manager
        A database manager instance.
    runner_kwargs
        Extra keyword argument to pass to the `adaptive.Runner`. Note that this dict
        will be serialized and pasted in the ``job_script``.
    goal
        The goal passed to the `adaptive.Runner`. Note that this function will
        be serialized and pasted in the ``job_script``. Can be a smart-goal
        that accepts
        ``Callable[[adaptive.BaseLearner], bool] | float | datetime | timedelta | None``.
        See `adaptive_scheduler.utils.smart_goal` for more information.
    log_interval
        Time in seconds between log entries.
    save_interval
        Time in seconds between saving of the learners.
    save_dataframe
        Whether to periodically save the learner's data as a `pandas.DataFame`.
    dataframe_format
        The format in which to save the `pandas.DataFame`. See the type hint for the options.
    loky_start_method
        Loky start method, by default "loky".

    Returns
    -------
    dict
        The command line options for the job_script.

    """
    if runner_kwargs is None:
        runner_kwargs = {}
    runner_kwargs["goal"] = goal
    base64_runner_kwargs = _serialize_to_b64(runner_kwargs)

    opts = {
        "--url": database_manager.url,
        "--log-interval": log_interval,
        "--save-interval": save_interval,
        "--serialized-runner-kwargs": base64_runner_kwargs,
    }
    if scheduler.single_job_script:
        # if `cores` or `executor_type` is a tuple then we set it
        # in `Scheduler.start_job`
        assert isinstance(scheduler.cores, int)
        n = scheduler.cores
        if scheduler.executor_type == "ipyparallel":
            n -= 1
        opts["--n"] = n
        opts["--executor-type"] = scheduler.executor_type

    opts["--loky-start-method"] = loky_start_method

    if save_dataframe:
        opts["--dataframe-format"] = dataframe_format
        opts["--save-dataframe"] = None  # type: ignore[assignment]
    return opts


class JobManager(BaseManager):
    """Job manager.

    Parameters
    ----------
    job_names
        List of unique names used for the jobs with the same length as
        `learners`. Note that a job name does not correspond to a certain
        specific learner.
    database_manager
        A `DatabaseManager` instance.
    scheduler
        A scheduler instance from `adaptive_scheduler.scheduler`.
    interval
        Time in seconds between checking and starting jobs.
    max_simultaneous_jobs
        Maximum number of simultaneously running jobs. By default no more than 500
        jobs will be running. Keep in mind that if you do not specify a ``runner.goal``,
        jobs will run forever, resulting in the jobs that were not initially started
        (because of this `max_simultaneous_jobs` condition) to not ever start.
    max_fails_per_job
        Maximum number of times that a job can fail. This is here as a fail switch
        because a job might fail instantly because of a bug inside your code.
        The job manager will stop when
        ``n_jobs * total_number_of_jobs_failed > max_fails_per_job`` is true.
    save_dataframe
        Whether to periodically save the learner's data as a `pandas.DataFame`.
    dataframe_format
        The format in which to save the `pandas.DataFame`. See the type hint for the options.
    loky_start_method
        Loky start method, by default "loky".
    log_interval
        Time in seconds between log entries.
    save_interval
        Time in seconds between saving of the learners.
    runner_kwargs
        Extra keyword argument to pass to the `adaptive.Runner`. Note that this dict
        will be serialized and pasted in the ``job_script``.
    goal
        The goal passed to the `adaptive.Runner`. Note that this function will
        be serialized and pasted in the ``job_script``. Can be a smart-goal
        that accepts
        ``Callable[[adaptive.BaseLearner], bool] | float | datetime | timedelta | None``.
        See `adaptive_scheduler.utils.smart_goal` for more information.

    Attributes
    ----------
    n_started : int
        Total number of jobs started by the `JobManager`.

    """

    def __init__(
        self,
        job_names: list[str],
        database_manager: DatabaseManager,
        scheduler: BaseScheduler,
        interval: float = 30,
        *,
        max_simultaneous_jobs: int = 100,
        max_fails_per_job: int = 50,
        # Command line launcher options
        save_dataframe: bool = True,
        dataframe_format: _DATAFRAME_FORMATS = "pickle",
        loky_start_method: LOKY_START_METHODS = "loky",
        log_interval: float = 60,
        save_interval: float = 300,
        runner_kwargs: dict[str, Any] | None = None,
        goal: GoalTypes | None = None,
    ) -> None:
        super().__init__()
        self.job_names = job_names
        self.database_manager = database_manager
        self.scheduler = scheduler
        self.interval = interval
        self.max_simultaneous_jobs = max_simultaneous_jobs
        self.max_fails_per_job = max_fails_per_job
        # Other attributes
        self.n_started = 0
        self._request_times: dict[str, str] = {}

        # Command line launcher options
        self.save_dataframe = save_dataframe
        self.dataframe_format = dataframe_format
        self.loky_start_method = loky_start_method
        self.log_interval = log_interval
        self.save_interval = save_interval
        self.runner_kwargs = runner_kwargs
        self.goal = goal

    @property
    def max_job_starts(self) -> int:
        """Equivalent to ``self.max_fails_per_job * len(self.job_names)``."""
        return self.max_fails_per_job * len(self.job_names)

    def _queued(self, queue: dict[str, dict[str, Any]]) -> set[str]:
        return {
            job["job_name"]
            for job in queue.values()
            if job["job_name"] in self.job_names
        }

    def _setup(self) -> None:
        name_prefix = self.job_names[0].rsplit("-", 1)[0]
        options = command_line_options(
            scheduler=self.scheduler,
            database_manager=self.database_manager,
            runner_kwargs=self.runner_kwargs,
            log_interval=self.log_interval,
            save_interval=self.save_interval,
            save_dataframe=self.save_dataframe,
            dataframe_format=self.dataframe_format,
            goal=self.goal,
            loky_start_method=self.loky_start_method,
        )
        # hack to get the options in the job_script  # noqa: FIX004
        self.scheduler._command_line_options = options
        if self.scheduler.single_job_script:
            self.scheduler.write_job_script(name_prefix, options, index=None)

    async def _update_database_and_get_not_queued(
        self,
    ) -> tuple[set[str], set[str]] | None:
        running = self.scheduler.queue(me_only=True)
        self.database_manager.update(running)  # in case some jobs died
        queued = self._queued(running)  # running `job_name`s
        not_queued = set(self.job_names) - queued
        n_done = self.database_manager.n_done()
        if n_done == len(self.job_names):
            return None  # we are finished!
        n_to_schedule = max(0, len(not_queued) - n_done)
        return queued, set(list(not_queued)[:n_to_schedule])

    async def _start_new_jobs(
        self,
        not_queued: set[str],
        queued: set[str],
        ex: ThreadPoolExecutor,
        loop: asyncio.AbstractEventLoop,
    ) -> None:
        num_jobs_to_start = min(
            len(not_queued),
            self.max_simultaneous_jobs - len(queued),
        )
        for _ in range(num_jobs_to_start):
            job_name = not_queued.pop()
            queued.add(job_name)
            index, fname = self.database_manager._choose_fname(job_name)
            log.debug(
                f"Starting `job_name={job_name}` with `index={index}` and `fname={fname}`",
            )
            await loop.run_in_executor(
                ex,
                partial(self.scheduler.start_job, name=job_name, index=index),
            )
            self.n_started += 1
            self._request_times[job_name] = _now()

    async def _manage(self) -> None:
        loop = asyncio.get_event_loop()
        with ThreadPoolExecutor() as ex:  # TODO: use asyncio.to_thread when Python≥3.9
            while True:
                try:
                    update = await self._update_database_and_get_not_queued()
                    if update is None:  # we are finished!
                        return
                    queued, not_queued = update
                    await self._start_new_jobs(not_queued, queued, ex, loop)
                    if self.n_started > self.max_job_starts:
                        msg = (
                            "Too many jobs failed, your Python code probably has a bug."
                        )
                        raise MaxRestartsReachedError(msg)  # noqa: TRY301
                    if await sleep_unless_task_is_done(
                        self.database_manager.task,  # type: ignore[arg-type]
                        self.interval,
                    ):  # if true, we are done
                        return
                except asyncio.CancelledError:  # noqa: PERF203
                    log.info("task was cancelled because of a CancelledError")
                    raise
                except MaxRestartsReachedError as e:
                    log.exception(
                        "too many jobs have failed, cancelling the job manager",
                        n_started=self.n_started,
                        max_fails_per_job=self.max_fails_per_job,
                        max_job_starts=self.max_job_starts,
                        exception=str(e),
                    )
                    raise
                except Exception as e:  # noqa: BLE001
                    log.exception("got exception when starting a job", exception=str(e))
                    if await sleep_unless_task_is_done(
                        self.database_manager.task,  # type: ignore[arg-type]
                        5,
                    ):  # if true, we are done
                        return



"""Tests for the common module of the server_support module."""

from __future__ import annotations

import asyncio
from pathlib import Path

import pytest

from adaptive_scheduler._server_support.common import (
    _get_all_files,
    _ipython_profiles,
    cleanup_scheduler_files,
    get_allowed_url,
    periodically_clean_ipython_profiles,
)

from .helpers import MockScheduler, temporary_working_directory


@pytest.mark.asyncio()
async def test_periodically_clean_ipython_profiles(
    mock_scheduler: MockScheduler,
) -> None:
    """Test that the periodic cleaning of IPython profiles works."""
    task = periodically_clean_ipython_profiles(
        mock_scheduler,
        interval=0.1,  # type: ignore[arg-type]
    )
    await asyncio.sleep(0.2)
    task.cancel()


def test_get_allowed_url() -> None:
    """Test that the allowed URL is a TCP URL."""
    url = get_allowed_url()
    assert url.startswith("tcp://")
    url2 = get_allowed_url()
    assert url != url2


def test_cleanup_scheduler_files(mock_scheduler: MockScheduler, tmp_path: Path) -> None:
    """Test that the scheduler files are cleaned up correctly."""
    job_names = ["job1", "job2"]
    for name in job_names:
        (tmp_path / f"{name}.mock").touch()
        (tmp_path / f"{name}-JOBID.out").touch()
        (tmp_path / f"{name}-JOBID.log").touch()

    assert len(list(tmp_path.glob("*"))) == 6
    move_to = tmp_path / "moved"
    with temporary_working_directory(tmp_path):
        cleanup_scheduler_files(job_names, mock_scheduler, move_to=move_to)
    assert len(list(tmp_path.glob("*"))) == 1
    assert len(list(move_to.glob("*"))) == 6


def test__get_all_files(mock_scheduler: MockScheduler, tmp_path: Path) -> None:
    """Test that all files are returned correctly."""
    job_names = ["job1", "job2"]
    for name in job_names:
        (tmp_path / f"{name}.mock").touch()
        (tmp_path / f"{name}-JOBID.out").touch()
        (tmp_path / f"{name}-JOBID.log").touch()
    with temporary_working_directory(tmp_path):
        all_files = _get_all_files(job_names, mock_scheduler)
        assert len(all_files) == 6


def test__ipython_profiles() -> None:
    """Test that the IPython profiles are returned correctly."""
    profiles = _ipython_profiles()
    assert isinstance(profiles, list)
    assert all(isinstance(p, Path) for p in profiles)



"""SLURM for Adaptive Scheduler."""

from __future__ import annotations

import copy
import getpass
import re
import subprocess
import textwrap
from distutils.spawn import find_executable
from functools import cached_property, lru_cache
from typing import TYPE_CHECKING, TypeVar

from adaptive_scheduler._scheduler.base_scheduler import BaseScheduler
from adaptive_scheduler._scheduler.common import run_submit

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Any

    from adaptive_scheduler.utils import EXECUTOR_TYPES


T = TypeVar("T")


def _maybe_as_tuple(
    x: T | tuple[T, ...] | None,
    n: int | None,
    *,
    check_type: type | None = None,
) -> tuple[T, ...] | T | None:
    if x is None:
        return x
    if check_type is not None and not isinstance(x, (check_type, tuple)):
        msg = f"Expected `{check_type}` or `tuple[{check_type}, ...]`, got `{type(x)}`"
        raise TypeError(msg)
    if n is None:
        return x
    if isinstance(x, tuple):
        assert len(x) == n
        return x
    return tuple(copy.deepcopy(x) for _ in range(n))


def _tuple_lengths(*maybe_tuple: tuple[Any, ...] | Any) -> int | None:
    """Get the length of the items that are in tuples."""
    length = None
    for y in maybe_tuple:
        if isinstance(y, tuple):
            if length is None:
                length = len(y)
            elif length != len(y):
                msg = "All tuples should have the same length."
                raise ValueError(msg)
    return length


class SLURM(BaseScheduler):
    """Base object for a Scheduler.

    ``cores``, ``nodes``, ``cores_per_node``, ``extra_scheduler``,
    ``executor_type``, ``extra_script``, ``exclusive``, ``extra_env_vars``,
    ``num_threads`` and ``partition`` can be either a single value or a tuple of
    values. If a tuple is given, then the length of the tuple should be the same
    as the number of learners (jobs) that are run. This allows for different
    resources for different jobs.

    Parameters
    ----------
    cores
        Number of cores per job (so per learner.)
        Either use `cores` or `nodes` and `cores_per_node`.
    nodes
        Number of nodes per job (so per learner.)
        Either `nodes` and `cores_per_node` or use `cores`.
    cores_per_node
        Number of cores per node.
        Either `nodes` and `cores_per_node` or use `cores`.
    partition
        The SLURM partition to submit the job to.
    exclusive
        Whether to use exclusive nodes (e.g., if SLURM it adds ``--exclusive`` as option).
    log_folder
        The folder in which to put the log-files.
    mpiexec_executable
        ``mpiexec`` executable. By default `mpiexec` will be
        used (so probably from ``conda``).
    executor_type
        The executor that is used, by default `concurrent.futures.ProcessPoolExecutor` is used.
        One can use ``"ipyparallel"``, ``"dask-mpi"``, ``"mpi4py"``,
        ``"loky"``, ``"sequential"``, or ``"process-pool"``.
    num_threads
        ``MKL_NUM_THREADS``, ``OPENBLAS_NUM_THREADS``, ``OMP_NUM_THREADS``, and
        ``NUMEXPR_NUM_THREADS`` will be set to this number.
    extra_scheduler
        Extra ``#SLURM`` (depending on scheduler type)
        arguments, e.g. ``["--exclusive=user", "--time=1"]`` or a tuple of lists,
        e.g. ``(["--time=10"], ["--time=20"]])`` for two jobs.
    extra_env_vars
        Extra environment variables that are exported in the job
        script. e.g. ``["TMPDIR='/scratch'", "PYTHONPATH='my_dir:$PYTHONPATH'"]``.
    extra_script
        Extra script that will be executed after any environment variables are set,
        but before the main scheduler is run.

    """

    # Attributes that all schedulers need to have
    _ext = ".sbatch"
    _submit_cmd = "sbatch"
    _JOB_ID_VARIABLE = "${SLURM_JOB_ID}"
    _options_flag = "SBATCH"
    _cancel_cmd = "scancel"

    def __init__(  # noqa: PLR0912, PLR0915, C901
        self,
        *,
        cores: int | tuple[int, ...] | None = None,
        nodes: int | tuple[int, ...] | None = None,
        cores_per_node: int | tuple[int, ...] | None = None,
        partition: str | tuple[str, ...] | None = None,
        exclusive: bool | tuple[bool, ...] = True,
        python_executable: str | None = None,
        log_folder: str | Path = "",
        mpiexec_executable: str | None = None,
        executor_type: EXECUTOR_TYPES | tuple[EXECUTOR_TYPES, ...] = "process-pool",
        num_threads: int | tuple[int, ...] = 1,
        extra_scheduler: list[str] | tuple[list[str], ...] | None = None,
        extra_env_vars: list[str] | tuple[list[str], ...] | None = None,
        extra_script: str | tuple[str, ...] | None = None,
        batch_folder: str | Path = "",
    ) -> None:
        """Initialize the scheduler."""
        # Store the original values
        self._cores = cores
        self._nodes = nodes
        self._cores_per_node = cores_per_node
        self._partition = partition
        self._executor_type = executor_type
        self._num_threads = num_threads
        self._exclusive = exclusive
        self.__extra_scheduler = extra_scheduler
        self.__extra_env_vars = extra_env_vars
        self.__extra_script = extra_script

        msg = "Specify either `nodes` and `cores_per_node`, or only `cores`, not both."
        if cores is None:
            if nodes is None or cores_per_node is None:
                raise ValueError(msg)
        elif nodes is not None or cores_per_node is not None:
            raise ValueError(msg)

        if extra_scheduler is None:
            extra_scheduler = []
        if extra_env_vars is None:
            extra_env_vars = []
        if extra_script is None:
            extra_script = ""

        # If any is a tuple, then all should be a tuple
        n = _tuple_lengths(
            cores,
            nodes,
            cores_per_node,
            partition,
            executor_type,
            num_threads,
            exclusive,
            extra_scheduler,
            extra_env_vars,
            extra_script,
        )
        single_job_script = n is None
        cores = _maybe_as_tuple(cores, n, check_type=int)
        self.nodes = nodes = _maybe_as_tuple(nodes, n, check_type=int)
        self.cores_per_node = _maybe_as_tuple(cores_per_node, n, check_type=int)
        self.partition = partition = _maybe_as_tuple(partition, n, check_type=str)
        executor_type = _maybe_as_tuple(executor_type, n, check_type=str)  # type: ignore[assignment]
        num_threads = _maybe_as_tuple(num_threads, n, check_type=int)  # type: ignore[assignment]
        self.exclusive = _maybe_as_tuple(exclusive, n, check_type=bool)
        extra_scheduler = _maybe_as_tuple(extra_scheduler, n, check_type=list)
        extra_env_vars = _maybe_as_tuple(extra_env_vars, n, check_type=list)
        extra_script = _maybe_as_tuple(extra_script, n, check_type=str)

        if self.cores_per_node is not None:
            if single_job_script:
                assert isinstance(self.cores_per_node, int)
                assert isinstance(nodes, int)
                assert isinstance(extra_scheduler, list)
                extra_scheduler.append(f"--ntasks-per-node={self.cores_per_node}")
                cores = self.cores_per_node * nodes
            else:
                assert isinstance(self.cores_per_node, tuple)
                assert isinstance(nodes, tuple)
                assert isinstance(extra_scheduler, tuple)
                for lst, cpn in zip(extra_scheduler, self.cores_per_node):
                    assert isinstance(lst, list)
                    lst.append(f"--ntasks-per-node={cpn}")
                cores = tuple(cpn * n for cpn, n in zip(self.cores_per_node, nodes))

        if partition is not None:
            if single_job_script:
                assert isinstance(partition, str)
                assert isinstance(extra_scheduler, list)
                if partition not in self.partitions:
                    msg = f"Invalid partition: {partition}, only {self.partitions} are available."
                    raise ValueError(msg)
                extra_scheduler.append(f"--partition={partition}")
            else:
                if any(p not in self.partitions for p in partition):
                    msg = f"Invalid partition: {partition}, only {self.partitions} are available."
                    raise ValueError(msg)
                assert isinstance(extra_scheduler, tuple)
                for lst, p in zip(extra_scheduler, partition):
                    assert isinstance(lst, list)
                    lst.append(f"--partition={p}")

        if single_job_script:
            assert isinstance(extra_scheduler, list)
            assert isinstance(exclusive, bool)
            if self.exclusive:
                extra_scheduler.append("--exclusive")
        else:
            assert isinstance(extra_scheduler, tuple)
            assert isinstance(self.exclusive, tuple)
            for _ex, lst in zip(self.exclusive, extra_scheduler):
                assert isinstance(lst, list)
                if _ex:
                    lst.append("--exclusive")

        assert cores is not None
        super().__init__(
            cores,
            python_executable=python_executable,
            log_folder=log_folder,
            mpiexec_executable=mpiexec_executable,
            executor_type=executor_type,
            num_threads=num_threads,
            extra_scheduler=extra_scheduler,
            extra_env_vars=extra_env_vars,
            extra_script=extra_script,
            batch_folder=batch_folder,
        )
        # SLURM specific
        self.mpiexec_executable = mpiexec_executable or "srun --mpi=pmi2"

    def __getstate__(self) -> dict[str, Any]:
        """Get the state of the SLURM scheduler."""
        state = super().__getstate__()
        state["cores"] = self._cores
        state["nodes"] = self._nodes
        state["cores_per_node"] = self._cores_per_node
        state["partition"] = self._partition
        state["executor_type"] = self._executor_type
        state["num_threads"] = self._num_threads
        state["exclusive"] = self._exclusive
        state["extra_scheduler"] = self.__extra_scheduler
        state["extra_env_vars"] = self.__extra_env_vars
        state["extra_script"] = self.__extra_script
        return state

    def __setstate__(self, state: dict[str, Any]) -> None:
        """Set the state of the SLURM scheduler."""
        self.__init__(**state)  # type: ignore[misc]

    def _ipyparallel(self, *, index: int | None = None) -> tuple[str, tuple[str, ...]]:
        cores = self._get_cores(index=index)
        job_id = self._JOB_ID_VARIABLE
        profile = "${profile}"
        # We need to reserve one core for the controller
        if self.nodes is not None and self.partition is not None and self.exclusive:
            if self.single_job_script:
                partition = self.partition
                nodes = self.nodes
            else:
                assert isinstance(self.partition, list)
                assert isinstance(self.nodes, list)
                assert index is not None
                partition = self.partition[index]
                nodes = self.nodes[index]
            assert isinstance(partition, str)
            assert isinstance(nodes, int)
            # Limit the number of cores to the maximum number of cores per node
            max_cores_per_node = self.partitions[partition]
            tot_cores = nodes * max_cores_per_node
            cores = min(cores, tot_cores - 1)
        else:  # noqa: PLR5501
            if self.single_job_script:
                assert isinstance(self.cores, int)
                cores = self.cores - 1
            else:
                assert isinstance(self.cores, tuple)
                assert index is not None
                cores = self.cores[index] - 1

        start = textwrap.dedent(
            f"""\
            profile=adaptive_scheduler_{job_id}

            echo "Creating profile {profile}"
            ipython profile create {profile}

            echo "Launching controller"
            ipcontroller --ip="*" --profile={profile} --log-to-file &
            sleep 10

            echo "Launching engines"
            srun --ntasks {cores} ipengine \\
                --profile={profile} \\
                --cluster-id='' \\
                --log-to-file &

            echo "Starting the Python script"
            srun --ntasks 1 {self.python_executable} {self.launcher} \\
            """,
        )
        custom = (f"    --profile {profile}",)
        return start, custom

    def job_script(self, options: dict[str, Any], *, index: int | None = None) -> str:
        """Get a jobscript in string form.

        Returns
        -------
        job_script
            A job script that can be submitted to SLURM.
        index
            The index of the job that is being run. This is used when
            specifying different resources for different jobs.

        """
        cores = self._get_cores(index=index)
        job_script = textwrap.dedent(
            f"""\
            #!/bin/bash
            #SBATCH --ntasks {cores}
            #SBATCH --no-requeue
            {{extra_scheduler}}

            {{extra_env_vars}}

            {{extra_script}}

            {{executor_specific}}
            """,
        )

        return job_script.format(
            extra_scheduler=self.extra_scheduler(index=index),
            extra_env_vars=self.extra_env_vars(index=index),
            extra_script=self.extra_script(index=index),
            executor_specific=self._executor_specific("${NAME}", options, index=index),
        )

    def start_job(self, name: str, *, index: int | None = None) -> None:
        """Writes a job script and submits it to the scheduler."""
        if self.single_job_script:
            name_prefix = name.rsplit("-", 1)[0]
        else:
            name_prefix = name
            assert index is not None
            options = self._multi_job_script_options(index)
            self.write_job_script(name_prefix, options, index=index)

        (output_fname,) = self.output_fnames(name)
        output_str = str(output_fname).replace(self._JOB_ID_VARIABLE, "%A")
        output_opt = f"--output {output_str}"
        name_opt = f"--job-name {name}"
        submit_cmd = (
            f"{self.submit_cmd} {name_opt} {output_opt} {self.batch_fname(name_prefix)}"
        )
        run_submit(submit_cmd, name)

    def queue(self, *, me_only: bool = True) -> dict[str, dict[str, str]]:
        """Get the queue of jobs."""
        python_format = {
            "JobID": 100,
            "Name": 100,
            "state": 100,
            "NumNodes": 100,
            "NumTasks": 100,
            "ReasonList": 4000,
            "SubmitTime": 100,
            "StartTime": 100,
            "UserName": 100,
            "Partition": 100,
        }  # (key -> length) mapping

        slurm_format = ",".join(f"{k}:{v}" for k, v in python_format.items())
        squeue_executable = find_executable("squeue")
        assert isinstance(squeue_executable, str)
        cmd = [
            squeue_executable,
            rf'--Format=",{slurm_format},"',
            "--noheader",
            "--array",
        ]
        if me_only:
            username = getpass.getuser()
            cmd.append(f"--user={username}")
        proc = subprocess.run(cmd, text=True, capture_output=True, check=False)
        output = proc.stdout

        if (
            "squeue: error" in output
            or "slurm_load_jobs error" in output
            or proc.returncode != 0
        ):
            msg = "SLURM is not responding."
            raise RuntimeError(msg)

        def line_to_dict(line: str) -> dict[str, str]:
            chars = list(line)
            info = {}
            for k, v in python_format.items():
                info[k] = "".join(chars[:v]).strip()
                chars = chars[v:]
            return info

        squeue = [line_to_dict(line) for line in output.split("\n")]
        states = ("PENDING", "RUNNING", "CONFIGURING")
        squeue = [info for info in squeue if info["state"] in states]
        running = {info.pop("JobID"): info for info in squeue}
        for info in running.values():
            info["job_name"] = info.pop("Name")
        return running

    @cached_property
    def partitions(self) -> dict[str, int]:
        """Get the partitions of the SLURM scheduler."""
        return slurm_partitions()  # type: ignore[return-value]


def _get_ncores(partition: str) -> int | None:
    numbers = re.findall(r"\d+", partition)
    if not numbers:
        return None
    return int(numbers[0])


@lru_cache(maxsize=1)
def slurm_partitions(
    *,
    timeout: int = 5,
    with_ncores: bool = True,
) -> list[str] | dict[str, int | None]:
    """Get the available slurm partitions, raises subprocess.TimeoutExpired after timeout."""
    output = subprocess.run(
        ["sinfo", "-ahO", "partition"],
        capture_output=True,
        timeout=timeout,
        check=False,
    )
    lines = output.stdout.decode("utf-8").split("\n")
    partitions = sorted(partition for line in lines if (partition := line.strip()))
    # Sort partitions alphabetically, but put the default partition first
    partitions = sorted(partitions, key=lambda s: ("*" not in s, s))
    # Remove asterisk, which is used for default partition
    partitions = [partition.replace("*", "") for partition in partitions]
    if not with_ncores:
        return partitions

    return {partition: _get_ncores(partition) for partition in partitions}



"""Scheduler classes for Adaptive Scheduler."""

from __future__ import annotations

import os
import os.path
import warnings
from distutils.spawn import find_executable

from adaptive_scheduler._scheduler.base_scheduler import BaseScheduler
from adaptive_scheduler._scheduler.local import LocalMockScheduler
from adaptive_scheduler._scheduler.pbs import PBS
from adaptive_scheduler._scheduler.slurm import SLURM, slurm_partitions

__all__ = [
    "LocalMockScheduler",
    "PBS",
    "SLURM",
    "slurm_partitions",
    "BaseScheduler",
    "DefaultScheduler",
]


def _get_default_scheduler() -> type[BaseScheduler]:
    """Determine which scheduler system is being used.

    It tries to determine it by running both PBS and SLURM commands.

    If both are available then one needs to set an environment variable
    called 'SCHEDULER_SYSTEM' which is either 'PBS' or 'SLURM'.

    For example add the following to your `.bashrc`

    ```bash
    export SCHEDULER_SYSTEM="PBS"
    ```

    By default it is "SLURM".
    """
    has_pbs = bool(find_executable("qsub")) and bool(find_executable("qstat"))
    has_slurm = bool(find_executable("sbatch")) and bool(find_executable("squeue"))

    default = SLURM
    default_msg = f"We set DefaultScheduler to '{default}'."
    scheduler_system = os.environ.get("SCHEDULER_SYSTEM", "").upper()
    if scheduler_system:
        if scheduler_system not in ("PBS", "SLURM"):
            warnings.warn(
                f"SCHEDULER_SYSTEM={scheduler_system} is not implemented."
                f"Use SLURM or PBS. {default_msg}",
                stacklevel=2,
            )
            return default
        return {"SLURM": SLURM, "PBS": PBS}[scheduler_system]  # type: ignore[return-value]
    if has_slurm and has_pbs:
        msg = f"Both SLURM and PBS are detected. {default_msg}"
        warnings.warn(msg, stacklevel=2)
        return default
    if has_pbs:
        return PBS
    if has_slurm:
        return SLURM
    msg = f"No scheduler system could be detected. {default_msg}"
    warnings.warn(msg, stacklevel=2)
    return default


DefaultScheduler = _get_default_scheduler()



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable

from adaptive_scheduler.scheduler import SLURM, slurm_partitions
from adaptive_scheduler.utils import _get_default_args

from .common import console
from .run_manager import RunManager

if TYPE_CHECKING:
    import adaptive

    from adaptive_scheduler.utils import _DATAFRAME_FORMATS, EXECUTOR_TYPES, GoalTypes


def slurm_run(
    learners: list[adaptive.BaseLearner],
    fnames: list[str] | list[Path],
    *,
    partition: str | tuple[str, ...] | None = None,
    nodes: int | tuple[int, ...] = 1,
    cores_per_node: int | tuple[int, ...] | None = None,
    goal: GoalTypes | None = None,
    folder: str | Path = "",
    name: str = "adaptive",
    num_threads: int | tuple[int, ...] = 1,
    save_interval: float = 300,
    log_interval: float = 300,
    cleanup_first: bool = True,
    save_dataframe: bool = True,
    dataframe_format: _DATAFRAME_FORMATS = "pickle",
    max_fails_per_job: int = 50,
    max_simultaneous_jobs: int = 100,
    exclusive: bool | tuple[bool, ...] = True,
    executor_type: EXECUTOR_TYPES | tuple[EXECUTOR_TYPES, ...] = "process-pool",
    extra_scheduler: list[str] | tuple[list[str], ...] | None = None,
    extra_run_manager_kwargs: dict[str, Any] | None = None,
    extra_scheduler_kwargs: dict[str, Any] | None = None,
    initializers: list[Callable[[], None]] | None = None,
) -> RunManager:
    """Run adaptive on a SLURM cluster.

    ``cores``, ``nodes``, ``cores_per_node``, ``extra_scheduler``,
    ``executor_type``, ``extra_script``, ``exclusive``, ``extra_env_vars``,
    ``num_threads`` and ``partition`` can be either a single value or a tuple of
    values. If a tuple is given, then the length of the tuple should be the same
    as the number of learners (jobs) that are run. This allows for different
    resources for different jobs.

    Parameters
    ----------
    learners
        A list of learners.
    fnames
        A list of filenames to save the learners.
    partition
        The partition to use. If None, then the default partition will be used.
        (The one marked with a * in `sinfo`). Use
        `adaptive_scheduler.scheduler.slurm_partitions` to see the
        available partitions.
    nodes
        The number of nodes to use.
    cores_per_node
        The number of cores per node to use. If None, then all cores on the partition
        will be used.
    goal
        The goal of the adaptive run. If None, then the run will continue
        indefinitely.
    folder
        The folder to save the adaptive_scheduler files such as logs, database,
        and ``.sbatch`` files in.
    name
        The name of the job.
    num_threads
        The number of threads to use.
    save_interval
        The interval at which to save the learners.
    log_interval
        The interval at which to log the status of the run.
    cleanup_first
        Whether to clean up the folder before starting the run.
    save_dataframe
        Whether to save the `pandas.DataFrame`s with the learners data.
    dataframe_format
        The format to save the `pandas.DataFrame`s in. See
        `adaptive_scheduler.utils.save_dataframes` for more information.
    max_fails_per_job
        The maximum number of times a job can fail before it is cancelled.
    max_simultaneous_jobs
        The maximum number of simultaneous jobs.
    executor_type
        The executor that is used, by default `concurrent.futures.ProcessPoolExecutor` is used.
        One can use ``"ipyparallel"``, ``"dask-mpi"``, ``"mpi4py"``,
        ``"loky"``, ``"sequential"``, or ``"process-pool"``.
    exclusive
        Whether to use exclusive nodes, adds ``"--exclusive"`` if True.
    extra_scheduler
        Extra ``#SLURM`` (depending on scheduler type)
        arguments, e.g. ``["--exclusive=user", "--time=1"]`` or a tuple of lists,
        e.g. ``(["--time=10"], ["--time=20"]])`` for two jobs.
    extra_run_manager_kwargs
        Extra keyword arguments to pass to the `RunManager`.
    extra_scheduler_kwargs
        Extra keyword arguments to pass to the `adaptive_scheduler.scheduler.SLURM`.
    initializers
        List of functions that are called before the job starts, can populate
        a cache.

    Returns
    -------
    RunManager

    """
    if partition is None:
        partitions = slurm_partitions()
        assert isinstance(partitions, dict)
        partition, ncores = next(iter(partitions.items()))
        console.log(
            f"Using default partition {partition} (The one marked"
            f" with a '*' in `sinfo`) with {ncores} cores."
            " Use `adaptive_scheduler.scheduler.slurm_partitions`"
            " to see the available partitions.",
        )
    if executor_type == "process-pool" and (
        nodes > 1 if isinstance(nodes, int) else any(n > 1 for n in nodes)
    ):
        msg = (
            "process-pool can maximally use a single node,"
            " use e.g., ipyparallel for multi node.",
        )
        raise ValueError(msg)
    folder = Path(folder)
    folder.mkdir(parents=True, exist_ok=True)
    if cores_per_node is None:
        partitions = slurm_partitions()
        assert isinstance(partitions, dict)
        cores_per_node = (
            tuple(partitions[p] for p in partition)  # type: ignore[misc]
            if isinstance(partition, tuple)
            else partitions[partition]
        )

    if extra_scheduler_kwargs is None:
        extra_scheduler_kwargs = {}
    if extra_scheduler is not None:
        # "extra_scheduler" used to be passed via the extra_scheduler_kwargs
        # this ensures backwards compatibility
        assert "extra_scheduler" not in extra_scheduler_kwargs
        extra_scheduler_kwargs["extra_scheduler"] = extra_scheduler

    slurm_kwargs = dict(
        _get_default_args(SLURM),
        nodes=nodes,
        cores_per_node=cores_per_node,
        partition=partition,
        log_folder=folder / "logs",
        batch_folder=folder / "batch_scripts",
        executor_type=executor_type,
        num_threads=num_threads,
        exclusive=exclusive,
        **extra_scheduler_kwargs,
    )
    scheduler = SLURM(**slurm_kwargs)
    # Below are the defaults for the RunManager
    kw = dict(
        _get_default_args(RunManager),
        scheduler=scheduler,
        learners=learners,
        fnames=fnames,
        goal=goal,
        save_interval=save_interval,
        log_interval=log_interval,
        move_old_logs_to=folder / "old_logs",
        db_fname=folder / f"{name}.db.json",
        job_name=name,
        cleanup_first=cleanup_first,
        save_dataframe=save_dataframe,
        dataframe_format=dataframe_format,
        max_fails_per_job=max_fails_per_job,
        max_simultaneous_jobs=max_simultaneous_jobs,
        initializers=initializers,
    )
    if extra_run_manager_kwargs is None:
        extra_run_manager_kwargs = {}
    return RunManager(**dict(kw, **extra_run_manager_kwargs))



"""Adaptive Scheduler notebook widgets."""

from __future__ import annotations

import asyncio
import os
from collections import defaultdict
from contextlib import contextmanager, suppress
from datetime import datetime, timedelta
from glob import glob
from pathlib import Path
from typing import TYPE_CHECKING, Generator

import numpy as np
import pandas as pd

from adaptive_scheduler.utils import load_dataframes

if TYPE_CHECKING:
    from typing import Any, Callable

    import ipywidgets as ipyw

    from adaptive_scheduler.scheduler import BaseScheduler
    from adaptive_scheduler.server_support import RunManager
    from adaptive_scheduler.utils import _DATAFRAME_FORMATS, FnamesTypes


def _get_fnames(run_manager: RunManager, *, only_running: bool) -> list[Path]:
    if only_running:
        fnames = []
        for entry in run_manager.database_manager.as_dicts():
            if entry["is_done"]:
                continue
            if entry["log_fname"] is not None:
                fnames.append(entry["log_fname"])
            fnames += entry["output_logs"]
        return sorted(map(Path, fnames))
    pattern = f"{run_manager.job_name}-*"
    logs = set(Path(run_manager.scheduler.log_folder).glob(pattern))
    logs |= set(Path().glob(pattern))
    return sorted(logs)


def _failed_job_logs(
    fnames: list[Path],
    run_manager: RunManager,
    *,
    only_running: bool,
) -> list[Path]:
    running = {
        Path(e["log_fname"]).stem
        for e in run_manager.database_manager.as_dicts()
        if e["log_fname"] is not None
    }
    fnames_set = {
        fname.stem for fname in fnames if fname.suffix != run_manager.scheduler.ext
    }
    failed_set = fnames_set - running
    failed = [Path(f) for stem in failed_set for f in glob(f"{stem}*")]  # noqa: PTH207

    def maybe_append(fname: str, other_dir: Path, lst: list[Path]) -> None:
        p = Path(fname)
        p_other = other_dir / p.name
        if p.exists():
            lst.append(p)
        elif p_other.exists():
            lst.append(p_other)

    if not only_running and run_manager.move_old_logs_to is not None:
        base = Path(run_manager.move_old_logs_to)
        for e in run_manager.database_manager.failed:
            if not e["is_done"]:
                for f in e["output_logs"]:
                    maybe_append(f, base, failed)
            maybe_append(e["log_fname"], base, failed)
    return failed


def _files_that_contain(fnames: list[Path], text: str) -> list[Path]:
    def contains(fname: Path, text: str) -> bool:
        with fname.open("r", encoding="utf-8") as f:
            return any(text in line for line in f)

    return [fname for fname in fnames if contains(fname, text)]


def _sort_fnames(
    sort_by: str,
    run_manager: RunManager,
    fnames: list[Path],
) -> list[Path] | list[tuple[str, Path]]:
    def _try(f: Callable[[Any], Any]) -> Callable[[Any], Any]:
        def _f(x: Any) -> Any:
            try:
                return f(x)
            except Exception:  # noqa: BLE001
                return x

        return _f

    def _sort_key(value: tuple[float | str, str]) -> tuple[float | int, str]:
        x, fname = value
        if isinstance(x, str):
            return -1, fname
        return float(x), fname

    def _vec_timedelta(ts: pd.Timestamp) -> str:
        now = np.datetime64(datetime.now())  # noqa: DTZ005
        dt = np.timedelta64(now - ts, "s")  # type: ignore[operator]
        return f"{dt} ago"

    mapping = {
        "Alphabetical": (None, lambda _: ""),
        "CPU %": ("cpu_usage", lambda x: f"{x:.1f}%"),
        "Mem %": ("mem_usage", lambda x: f"{x:.1f}%"),
        "Last editted": ("timestamp", _vec_timedelta),
        "Loss": ("latest_loss", lambda x: f"{x:.2e}"),
        "npoints": ("npoints", lambda x: f"{x} pnts"),
        "Elapsed time": ("elapsed_time", lambda x: f"{x / 1e9}s"),
    }

    def extract(df: pd.DataFrame, fname: Path, key: str) -> Any | str:
        df_sel = df[df.log_fname.str.contains(fname.name)]
        values = df_sel[key].to_numpy()
        if values:
            return values[0]
        return "?"

    if sort_by != "Alphabetical":
        fname_mapping = defaultdict(list)
        for fname in fnames:
            fname_mapping[fname.stem].append(fname)

        df = run_manager.parse_log_files()
        if df.empty:
            return fnames
        log_fnames = set(df.log_fname.apply(Path))  # type: ignore [arg-type]
        df_key, transform = mapping[sort_by]
        assert df_key is not None  # for mypy
        stems = [fname.stem for fname in log_fnames]
        vals = [extract(df, fname, df_key) for fname in log_fnames]
        val_stem = sorted(zip(vals, stems), key=_sort_key, reverse=True)

        result: list[tuple[str, Path]] = []
        for val, stem in val_stem:
            val = _try(transform)(val)  # noqa: PLW2901
            for fname in fname_mapping[stem]:
                result.append((f"{val}: {fname.name}", fname))  # noqa: PERF401

        missing = fname_mapping.keys() - set(stems)
        for stem in sorted(missing):
            for fname in fname_mapping[stem]:
                result.append((f"?: {fname.name}", fname))  # noqa: PERF401
        return result

    return fnames


def _read_file(fname: Path, max_lines: int = 500) -> str:
    try:
        with fname.open("r", encoding="utf-8") as f:
            lines = f.readlines()

            if len(lines) > max_lines:
                lines = lines[-max_lines:]
                lines.insert(0, f"Only displaying the last {max_lines} lines!")
                lines.insert(1, "\n")
            return "".join(lines)
    except UnicodeDecodeError:
        return f"Could not decode file ({fname})!"
    except Exception as e:  # noqa: BLE001
        return f"Exception with trying to read {fname}:\n{e}."


def log_explorer(run_manager: RunManager) -> ipyw.VBox:  # noqa: C901, PLR0915
    """Log explorer widget."""
    import ipywidgets as ipyw

    def _update_fname_dropdown(
        run_manager: RunManager,
        fname_dropdown: ipyw.Dropdown,
        only_running_checkbox: ipyw.Checkbox,
        only_failed_checkbox: ipyw.Checkbox,
        sort_by_dropdown: ipyw.Dropdown,
        contains_text: ipyw.Text,
    ) -> Callable[[Any], None]:
        def on_click(_: Any) -> None:
            current_value = fname_dropdown.value
            fnames = _get_fnames(run_manager, only_running=only_running_checkbox.value)
            if only_failed_checkbox.value:
                fnames = _failed_job_logs(
                    fnames,
                    run_manager,
                    only_running=only_running_checkbox.value,
                )
            if contains_text.value.strip():
                fnames = _files_that_contain(fnames, contains_text.value.strip())
            sorted_fnames = _sort_fnames(sort_by_dropdown.value, run_manager, fnames)
            fname_dropdown.options = sorted_fnames
            with suppress(Exception):
                fname_dropdown.value = current_value
            fname_dropdown.disabled = not sorted_fnames

        return on_click

    def _last_editted(fname: Path) -> float:
        try:
            return fname.stat().st_mtime
        except FileNotFoundError:
            return -1.0

    async def _tail_log(fname: Path, textarea: ipyw.Textarea) -> None:
        t = -2.0  # to make sure the update always triggers
        while True:
            await asyncio.sleep(2)
            try:
                t_new = _last_editted(fname)
                if t_new > t:
                    textarea.value = _read_file(fname, run_manager.max_log_lines)
                    t = t_new
            except asyncio.CancelledError:
                return
            except Exception:  # noqa: S110, BLE001
                pass

    def _tail(
        dropdown: ipyw.Dropdown,
        tail_button: ipyw.Button,
        textarea: ipyw.Textarea,
        update_button: ipyw.Button,
        only_running_checkbox: ipyw.Checkbox,
        only_failed_checkbox: ipyw.Checkbox,
        sort_by_dropdown: ipyw.Dropdown,
        contains_text: ipyw.Text,
    ) -> Callable[[Any], None]:
        tail_task = None
        ioloop = asyncio.get_running_loop()

        def on_click(_: Any) -> None:
            nonlocal tail_task
            tailing_log = tail_task is not None

            def update_ui_state(tailing: bool) -> None:  # noqa: FBT001
                tail_button.description = "cancel tail log" if tailing else "tail log"
                tail_button.button_style = "danger" if tailing else "info"
                tail_button.icon = "window-close" if tailing else "refresh"
                dropdown.disabled = tailing
                only_running_checkbox.disabled = tailing
                only_failed_checkbox.disabled = tailing
                update_button.disabled = tailing
                sort_by_dropdown.disabled = tailing
                contains_text.disabled = tailing

            if not tailing_log:
                fname = dropdown.options[dropdown.index]
                tail_task = ioloop.create_task(_tail_log(fname, textarea))
            else:
                assert tail_task is not None
                tail_task.cancel()
                tail_task = None

            update_ui_state(not tailing_log)

        return on_click

    def _on_dropdown_change(
        textarea: ipyw.Textarea,
    ) -> Callable[[dict[str, Any]], None]:
        def on_change(change: dict[str, Any]) -> None:
            if (
                change["type"] == "change"
                and change["name"] == "value"
                and change["new"] is not None
            ):
                textarea.value = _read_file(change["new"], run_manager.max_log_lines)

        return on_change

    def _click_button_on_change(
        button: ipyw.Button,
    ) -> Callable[[dict[str, Any]], None]:
        def on_change(change: dict[str, Any]) -> None:
            if change["type"] == "change" and change["name"] == "value":
                button.click()

        return on_change

    fnames = _get_fnames(run_manager, only_running=False)
    # no need to sort `fnames` because the default sort_by option is alphabetical
    text = _read_file(fnames[0], run_manager.max_log_lines) if fnames else ""
    textarea = ipyw.Textarea(text, layout={"width": "auto"}, rows=20)
    sort_by_dropdown = ipyw.Dropdown(
        description="Sort by",
        options=["Alphabetical", "CPU %", "Mem %", "Last editted", "Loss", "npoints"],
    )
    contains_text = ipyw.Text(description="Has string")
    fname_dropdown = ipyw.Dropdown(description="File name", options=fnames)
    fname_dropdown.observe(_on_dropdown_change(textarea))
    only_running_checkbox = ipyw.Checkbox(
        description="Only files of running jobs",
        indent=False,
    )
    only_failed_checkbox = ipyw.Checkbox(
        description="Only files of failed jobs (might include false positives)",
        indent=False,
    )
    update_button = ipyw.Button(
        description="update file list",
        button_style="info",
        icon="refresh",
    )
    update_button.on_click(
        _update_fname_dropdown(
            run_manager,
            fname_dropdown,
            only_running_checkbox,
            only_failed_checkbox,
            sort_by_dropdown,
            contains_text,
        ),
    )
    sort_by_dropdown.observe(_click_button_on_change(update_button))
    only_running_checkbox.observe(_click_button_on_change(update_button))
    only_failed_checkbox.observe(_click_button_on_change(update_button))
    tail_button = ipyw.Button(
        description="tail log",
        button_style="info",
        icon="refresh",
    )
    tail_button.on_click(
        _tail(
            fname_dropdown,
            tail_button,
            textarea,
            update_button,
            only_running_checkbox,
            only_failed_checkbox,
            sort_by_dropdown,
            contains_text,
        ),
    )
    vbox = ipyw.VBox(
        [
            only_running_checkbox,
            only_failed_checkbox,
            update_button,
            sort_by_dropdown,
            contains_text,
            fname_dropdown,
            tail_button,
            textarea,
        ],
        layout=ipyw.Layout(border="solid 2px gray"),
    )
    _add_title("adaptive_scheduler.widgets.log_explorer", vbox)
    return vbox


def _add_title(title: str, vbox: ipyw.VBox) -> None:
    import ipywidgets as ipyw

    title = ipyw.HTML(f"<h2><tt>{title}</tt></h2>")
    vbox.children = (title, *vbox.children)


def _bytes_to_human_readable(size_in_bytes: int) -> str:
    if size_in_bytes < 0:
        msg = "Size must be a positive integer"
        raise ValueError(msg)

    units = ["B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB"]
    index = 0
    bytes_in_kb = 1024
    while size_in_bytes >= bytes_in_kb and index < len(units) - 1:
        size_in_bytes /= bytes_in_kb  # type: ignore[assignment]
        index += 1

    return f"{size_in_bytes:.2f} {units[index]}"


def _timedelta_to_human_readable(
    time_input: timedelta | float,
    *,
    short_format: bool = True,
) -> str:
    """Convert a timedelta object or an int (in seconds) into a human-readable format."""
    if isinstance(time_input, timedelta):
        total_seconds = int(time_input.total_seconds())
    elif isinstance(time_input, (int, float)):
        total_seconds = int(time_input)
    else:
        msg = "Input must be a datetime.timedelta object or an int (in seconds)"
        raise TypeError(msg)

    days, remainder = divmod(total_seconds, 86400)
    hours, remainder = divmod(remainder, 3600)
    minutes, seconds = divmod(remainder, 60)

    periods = [
        ("day", days),
        ("hour", hours),
        ("minute", minutes),
        ("second", seconds),
    ]

    result = []
    for period, value in periods:
        if value:
            period_label = period if value == 1 else f"{period}s"
            if short_format:
                period_label = period_label[0]
            result.append(f"{value:.0f} {period_label}")

    return ", ".join(result)


def _total_size(fnames: FnamesTypes) -> int:
    """Return the total size of the files in `fnames`."""

    def flatten(
        items: FnamesTypes,
    ) -> Generator[str | Path, None, None]:
        """Flatten nested lists."""
        for item in items:
            if isinstance(item, (list, tuple)):
                yield from flatten(item)
            else:
                yield item

    flattened_fnames = list(flatten(fnames))
    return sum(
        os.path.getsize(str(fname))  # noqa: PTH202
        for fname in flattened_fnames
        if os.path.isfile(fname)  # noqa: PTH113
    )


def _interp_red_green(
    percent: float,
    pct_red: int = 30,
    pct_green: int = 10,
) -> tuple[int, int, int]:
    if pct_green < pct_red:
        if percent <= pct_green:
            return 0, 255, 0
        if percent >= pct_red:
            return 255, 0, 0
    else:
        if percent >= pct_green:
            return 0, 255, 0
        if percent <= pct_red:
            return 255, 0, 0

    # Interpolate between green and red
    factor = (percent - pct_green) / (pct_red - pct_green)
    red_level = int(255 * factor)
    green_level = int(255 * (1 - factor))
    return red_level, green_level, 0


def _create_html_tag(value: float, color: tuple[int, int, int]) -> str:
    red_level, green_level, blue_level = color
    hex_color = f"#{red_level:02x}{green_level:02x}{blue_level:02x}"
    return f'<font color="{hex_color}">{value:.2f}%</font>'


def _info_html(run_manager: RunManager) -> str:
    queue = run_manager.scheduler.queue(me_only=True)
    dbm = run_manager.database_manager
    dbm.update(queue)
    jobs = [job for job in queue.values() if job["job_name"] in run_manager.job_names]
    n_running = sum(job["state"] in ("RUNNING", "R") for job in jobs)
    n_pending = sum(job["state"] in ("PENDING", "Q", "CONFIGURING") for job in jobs)
    n_done = sum(1 for job in dbm.as_dicts() if job["is_done"])
    n_failed = len(dbm.failed)
    n_failed_color = "red" if n_failed > 0 else "black"

    status = run_manager.status()
    color = {
        "cancelled": "orange",
        "not yet started": "orange",
        "running": "blue",
        "failed": "red",
        "finished": "green",
    }[status]

    def _table_row(i: int, key: str, value: Any) -> str:
        """Style the rows of a table. Based on the default Jupyterlab table style."""
        style = "text-align: right; padding: 0.5em 0.5em; line-height: 1.0;"
        if i % 2 == 1:
            style += " background: var(--md-grey-100);"
        return (
            f'<tr><th style="{style}">{key}</th><th style="{style}">{value}</th></tr>'
        )

    data_size = _total_size(dbm.fnames)
    info = [
        ("status", f'<font color="{color}">{status}</font>'),
        ("# running jobs", f'<font color="blue">{n_running}</font>'),
        ("# pending jobs", f'<font color="orange">{n_pending}</font>'),
        ("# finished jobs", f'<font color="green">{n_done}</font>'),
        ("# failed jobs", f'<font color="{n_failed_color}">{n_failed}</font>'),
        ("elapsed time", timedelta(seconds=run_manager.elapsed_time())),
        ("total data size", _bytes_to_human_readable(data_size)),
    ]
    if dbm._total_learner_size is not None:
        info.append(
            ("empty learner size", _bytes_to_human_readable(dbm._total_learner_size)),
        )

    starting_times = run_manager.job_starting_times()
    if starting_times:
        mean_starting_time = _timedelta_to_human_readable(np.mean(starting_times))  # type: ignore[arg-type]
        std_starting_time = _timedelta_to_human_readable(np.std(starting_times))  # type: ignore[arg-type]
        info.append(("avg job start time", mean_starting_time))
        info.append(("std job start time", std_starting_time))

    with suppress(Exception):
        df = run_manager.parse_log_files()
        t_last = (pd.Timestamp.now() - df.timestamp.max()).seconds

        cpu = df.cpu_usage.mean()
        cpu_html_value = _create_html_tag(cpu, _interp_red_green(cpu, 50, 80))

        mem = df.mem_usage.mean()
        mem_html_value = _create_html_tag(mem, _interp_red_green(mem, 80, 50))

        max_mem = df.mem_usage.max()
        max_mem_html_value = _create_html_tag(
            max_mem,
            _interp_red_green(max_mem, 80, 50),
        )

        overhead = df.overhead.mean()
        overhead_html_value = _create_html_tag(
            overhead,
            _interp_red_green(overhead, 10, 30),
        )

        from_logs = [
            ("# of points", df.npoints.sum()),
            ("mean CPU usage", cpu_html_value),
            ("mean memory usage", mem_html_value),
            ("max memory usage", max_mem_html_value),
            ("mean overhead", overhead_html_value),
            ("last log-entry", f"{t_last}s ago"),
        ]
        for key in ["npoints/s", "latest_loss", "nlearners"]:
            with suppress(Exception):
                from_logs.append((f"mean {key}", f"{df[key].mean():.1f}"))
        msg = "this is extracted from the log files, so it might not be up-to-date"
        abbr = '<abbr title="{}">{}</abbr>'  # creates a tooltip
        info.extend([(abbr.format(msg, k), v) for k, v in from_logs])

    table = "\n".join(_table_row(i, k, v) for i, (k, v) in enumerate(info))

    return f"""
        <table>
        {table}
        </table>
    """


def _create_widget(
    data_provider: Callable[[], pd.DataFrame],
    update_button_text: str,
    *,
    itables_checkbox_default: bool = False,
    additional_widgets: list[ipyw.Widget] | None = None,
    extra_widget_config: dict | None = None,
) -> tuple[ipyw.VBox, Callable[[Any], None]]:
    import ipywidgets as ipyw
    from IPython.display import display

    def _update_data_df(
        itables_checkbox: ipyw.Checkbox,
        output_widget: ipyw.Output,
    ) -> Callable[[Any], None]:
        def on_click(_: Any) -> None:
            with output_widget:
                output_widget.clear_output()
                df = data_provider()
                if itables_checkbox.value:
                    from itables import show

                    _set_itables_opts()
                    show(df)
                else:
                    with _display_all_dataframe_rows():
                        display(df)

        return on_click

    # Create widgets
    output_widget = ipyw.Output()
    itables_checkbox = ipyw.Checkbox(
        description="Use itables (interactive)",
        indent=False,
        value=itables_checkbox_default,
    )
    update_button = ipyw.Button(
        description=update_button_text,
        button_style="info",
        icon="refresh",
    )

    # Update the DataFrame in the ipyw.Output widget when the button is clicked or the checkbox is changed
    update_function = _update_data_df(
        itables_checkbox,
        output_widget,
    )
    update_button.on_click(update_function)
    itables_checkbox.observe(update_function, names="value")

    # Initialize the DataFrame display
    update_function(None)

    # Create a ipyw.VBox and add the widgets to it
    widget_list = [itables_checkbox, update_button, output_widget]
    if additional_widgets:
        widget_list = additional_widgets + widget_list

    if extra_widget_config:
        for config in extra_widget_config.values():
            widget_list.insert(config["position"], config["widget"])

    vbox = ipyw.VBox(widget_list, layout=ipyw.Layout(border="solid 2px gray"))
    return (vbox, update_function)


def queue_widget(scheduler: BaseScheduler) -> ipyw.VBox:
    """Create a widget that shows the current queue and allows to update it."""
    import ipywidgets as ipyw

    me_only_checkbox = ipyw.Checkbox(
        description="Only my jobs",
        indent=False,
        value=True,
    )

    def get_queue_df() -> pd.DataFrame:
        queue = scheduler.queue(me_only=me_only_checkbox.value)
        return pd.DataFrame(queue).transpose()

    # Get both the VBox and the update_function from _create_widget
    vbox, update_function = _create_widget(
        get_queue_df,
        "Update queue",
        additional_widgets=[me_only_checkbox],
    )

    # Add an observer to the 'me_only_checkbox' that calls the 'update_function' when the checkbox value changes
    me_only_checkbox.observe(update_function, names="value")
    _add_title("adaptive_scheduler.widgets.queue_widget", vbox)
    return vbox


def database_widget(run_manager: RunManager) -> ipyw.VBox:
    """Create a widget that shows the current database and allows to update it."""

    def get_database_df() -> pd.DataFrame:
        return run_manager.database_manager.as_df()

    vbox, _ = _create_widget(get_database_df, "Update database")
    _add_title("adaptive_scheduler.widgets.database_widget", vbox)
    return vbox


def _set_itables_opts() -> None:
    import itables.options as opt

    opt.maxBytes = 262_144


def results_widget(
    fnames: list[str] | list[Path],
    dataframe_format: _DATAFRAME_FORMATS,
) -> ipyw.VBox:
    """Widget that loads and displays the results as `pandas.DataFrame`s."""
    import ipywidgets as ipyw

    def on_concat_checkbox_value_change(change: dict) -> None:
        if change["name"] == "value":
            dropdown.layout.visibility = "hidden" if change["new"] else "visible"
            update_function(None)

    def get_results_df() -> pd.DataFrame:
        selected_fname = dropdown.value
        dfs = [selected_fname] if not concat_checkbox.value else fnames
        df = load_dataframes(dfs, format=dataframe_format)
        assert isinstance(df, pd.DataFrame)

        if len(df) > max_rows.value:
            sample_indices: np.ndarray = np.linspace(
                0,
                len(df) - 1,
                num=max_rows.value,
                dtype=int,
            )
            df = df.iloc[sample_indices]

        return df  # type: ignore[return-value]

    # Create widgets
    dropdown = ipyw.Dropdown(options=fnames)
    concat_checkbox = ipyw.Checkbox(description="Concat all dataframes", indent=False)
    max_rows = ipyw.IntText(value=300, description="Max rows")

    # Observe the value change in the 'concat_checkbox'
    concat_checkbox.observe(on_concat_checkbox_value_change, names="value")

    extra_widget_config = {
        "concat_checkbox": {"widget": concat_checkbox, "position": 1},
        "dropdown": {"widget": dropdown, "position": 0},
        "max_rows": {"widget": max_rows, "position": 4},
    }

    vbox, update_function = _create_widget(
        get_results_df,
        "Update results",
        extra_widget_config=extra_widget_config,
    )

    # Add observers for the 'dropdown' and 'max_rows' widgets
    dropdown.observe(update_function, names="value")
    max_rows.observe(update_function, names="value")

    _add_title("adaptive_scheduler.widgets.results_widget", vbox)
    return vbox


def _toggle_widget(
    widget_key: str,
    widget_dict: dict[str, ipyw.Widget | str],
    toggle_dict: dict[str, dict[str, Any]],
) -> Callable[[Any], None]:
    import ipywidgets as ipyw
    from IPython.display import display

    def on_click(_: Any) -> None:
        widget = toggle_dict[widget_key]["widget"]
        if widget is None:
            widget = toggle_dict[widget_key]["init_func"]()
            toggle_dict[widget_key]["widget"] = widget

        button = widget_dict[widget_key]
        assert isinstance(button, ipyw.Button)
        show_description = f"show {widget_key}"
        hide_description = f"hide {widget_key}"
        output = toggle_dict[widget_key]["output"]
        if button.description == show_description:
            button.description = hide_description
            button.button_style = "warning"
            with output:
                output.clear_output()
                display(widget)
        else:
            button.description = show_description
            button.button_style = "info"
            with output:
                output.clear_output()

    return on_click


def _switch_to(
    box: ipyw.HBox,
    *buttons: ipyw.Button,
    _callable: Callable[[], None] | None = None,
) -> Callable[[Any], None]:
    def on_click(_: Any) -> None:
        box.children = tuple(buttons)
        if _callable is not None:
            _callable()

    return on_click


def _create_confirm_deny(
    initial_button: ipyw.Button,
    widgets: dict[str, ipyw.Button | ipyw.HBox],
    callable_func: Callable[[], None],
    key: str,
) -> None:
    import ipywidgets as ipyw

    confirm_button = ipyw.Button(
        description="Confirm",
        button_style="success",
        icon="check",
    )
    deny_button = ipyw.Button(
        description="Deny",
        button_style="danger",
        icon="close",
    )

    initial_button.on_click(
        _switch_to(widgets[key], confirm_button, deny_button),
    )
    deny_button.on_click(_switch_to(widgets[key], initial_button))
    confirm_button.on_click(
        _switch_to(widgets[key], initial_button, _callable=callable_func),
    )


def info(run_manager: RunManager) -> None:
    """Display information about the `RunManager`.

    Returns an interactive ipywidget that can be
    visualized in a Jupyter notebook.
    """
    import ipywidgets as ipyw
    from IPython.display import display

    _disable_widgets_output_scrollbar()

    status = ipyw.HTML(value=_info_html(run_manager))

    layout = ipyw.Layout(width="200px")

    cancel_button = ipyw.Button(
        description="cancel jobs",
        layout=layout,
        button_style="danger",
        icon="stop",
    )
    cleanup_button = ipyw.Button(
        description="cleanup log and batch files",
        layout=layout,
        button_style="danger",
        icon="remove",
    )
    update_info_button = ipyw.Button(
        description="update info",
        layout=layout,
        icon="refresh",
    )
    update_info_button.style.button_color = "lightgreen"
    load_learners_button = ipyw.Button(
        description="load learners",
        layout=layout,
        button_style="info",
        icon="download",
    )
    show_logs_button = ipyw.Button(
        description="show logs",
        layout=layout,
        button_style="info",
        icon="book",
    )
    show_queue_button = ipyw.Button(
        description="show queue",
        layout=layout,
        button_style="info",
        icon="tasks",
    )
    show_db_button = ipyw.Button(
        description="show database",
        layout=layout,
        button_style="info",
        icon="database",
    )
    show_results_button = ipyw.Button(
        description="show results",
        layout=layout,
        button_style="info",
        icon="table",
    )
    widgets = {
        "update info": update_info_button,
        "cancel": ipyw.HBox([cancel_button], layout=layout),
        "cleanup": ipyw.HBox([cleanup_button], layout=layout),
        "load learners": load_learners_button,
        "logs": show_logs_button,
        "queue": show_queue_button,
        "database": show_db_button,
        "results": show_results_button,
    }

    def update(_: Any) -> None:
        status.value = _info_html(run_manager)

    def load_learners(_: Any) -> None:
        run_manager.load_learners()

    toggle_dict = {
        "logs": {
            "widget": None,
            "init_func": lambda: log_explorer(run_manager),
            "show_description": "logs",
            "hide_description": "hide logs",
            "output": ipyw.Output(),
        },
        "queue": {
            "widget": None,
            "init_func": lambda: queue_widget(run_manager.scheduler),
            "output": ipyw.Output(),
        },
        "database": {
            "widget": None,
            "init_func": lambda: database_widget(run_manager),
            "output": ipyw.Output(),
        },
        "results": {
            "widget": None,
            "init_func": lambda: results_widget(
                run_manager.fnames,
                run_manager.dataframe_format,
            ),
            "output": ipyw.Output(),
        },
    }

    def cancel() -> None:
        run_manager.cancel()
        update(None)

    def cleanup(*, include_old_logs: ipyw.Checkbox) -> Callable[[], None]:
        def _callable() -> None:
            run_manager.cleanup(remove_old_logs_folder=include_old_logs.value)
            update(None)

        return _callable

    widgets["update info"].on_click(update)
    toggle_logs = _toggle_widget("logs", widgets, toggle_dict)
    toggle_queue = _toggle_widget("queue", widgets, toggle_dict)
    toggle_database = _toggle_widget("database", widgets, toggle_dict)
    toggle_results = _toggle_widget("results", widgets, toggle_dict)
    widgets["logs"].on_click(toggle_logs)
    widgets["queue"].on_click(toggle_queue)
    widgets["database"].on_click(toggle_database)
    widgets["results"].on_click(toggle_results)
    widgets["load learners"].on_click(load_learners)

    # Cancel button with confirm/deny option
    _create_confirm_deny(cancel_button, widgets, cancel, key="cancel")

    # Cleanup button with confirm/deny option
    include_old_logs = ipyw.Checkbox(
        value=False,
        description=f"Remove {run_manager.move_old_logs_to}/ folder",
        indent=False,
    )
    # Cleanup button with confirm/deny option
    cleanup_callable = cleanup(include_old_logs=include_old_logs)
    _create_confirm_deny(cleanup_button, widgets, cleanup_callable, key="cleanup")
    buttons_box = ipyw.VBox(tuple(widgets.values()))
    buttons_box.layout.margin = "0 0 0 100px"
    top_box = ipyw.HBox((status, buttons_box))
    box = ipyw.VBox((top_box, *(v["output"] for v in toggle_dict.values())))
    display(box)


@contextmanager
def _display_all_dataframe_rows(max_colwidth: int = 50) -> Generator[None, None, None]:
    """Display all rows in a `pandas.DataFrame`."""
    original_max_rows = pd.options.display.max_rows
    original_max_colwidth = pd.options.display.max_colwidth
    pd.set_option("display.max_rows", None)
    pd.set_option("display.max_colwidth", max_colwidth)
    try:
        yield
    finally:
        pd.set_option("display.max_rows", original_max_rows)
        pd.set_option("display.max_colwidth", original_max_colwidth)


def _disable_widgets_output_scrollbar() -> None:
    import ipywidgets as ipyw
    from IPython.display import display

    style = """
        <style>
            .jupyter-widgets-output-area .output_scroll {
                height: unset !important;
                border-radius: unset !important;
                -webkit-box-shadow: unset !important;
                box-shadow: unset !important;
            }
            .jupyter-widgets-output-area  {
                height: auto !important;
            }
        </style>
        """
    display(ipyw.HTML(style))



"""LocalMockScheduler for Adaptive Scheduler."""

from __future__ import annotations

import textwrap
from typing import TYPE_CHECKING

from adaptive_scheduler._scheduler.base_scheduler import BaseScheduler
from adaptive_scheduler._scheduler.common import run_submit

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Any

    from adaptive_scheduler.utils import EXECUTOR_TYPES


class LocalMockScheduler(BaseScheduler):
    """A scheduler that can be used for testing and runs locally."""

    # Attributes that all schedulers need to have
    _ext = ".batch"
    _JOB_ID_VARIABLE = "${JOB_ID}"

    def __init__(
        self,
        cores: int,
        *,
        python_executable: str | None = None,
        log_folder: str | Path = "",
        mpiexec_executable: str | None = None,
        executor_type: EXECUTOR_TYPES = "process-pool",
        num_threads: int = 1,
        extra_scheduler: list[str] | tuple[list[str], ...] | None = None,
        extra_env_vars: list[str] | None = None,
        extra_script: str | None = None,
        batch_folder: str | Path = "",
        # LocalMockScheduler specific
        mock_scheduler_kwargs: dict[str, Any] | None = None,
    ) -> None:
        """Initialize the LocalMockScheduler."""
        import adaptive_scheduler._mock_scheduler

        super().__init__(
            cores,
            python_executable=python_executable,
            log_folder=log_folder,
            mpiexec_executable=mpiexec_executable,
            executor_type=executor_type,
            num_threads=num_threads,
            extra_scheduler=extra_scheduler,
            extra_env_vars=extra_env_vars,
            extra_script=extra_script,
            batch_folder=batch_folder,
        )
        # LocalMockScheduler specific
        self.mock_scheduler_kwargs = mock_scheduler_kwargs or {}
        self.mock_scheduler = adaptive_scheduler._mock_scheduler.MockScheduler(
            **self.mock_scheduler_kwargs,
        )
        mock_scheduler_file = adaptive_scheduler._mock_scheduler.__file__
        self.base_cmd = f"{self.python_executable} {mock_scheduler_file}"

        # Attributes that all schedulers need to have
        self._submit_cmd = f"{self.base_cmd} --submit"  # type: ignore[misc]
        self._cancel_cmd = f"{self.base_cmd} --cancel"  # type: ignore[misc]

    def __getstate__(self) -> dict[str, Any]:
        """Get the state of the scheduler."""
        # LocalMockScheduler has one different argument from the BaseScheduler
        return dict(
            **super().__getstate__(),
            mock_scheduler_kwargs=self.mock_scheduler_kwargs,
        )

    def job_script(self, options: dict[str, Any], *, index: int | None = None) -> str:
        """Get a jobscript in string form.

        Returns
        -------
        job_script
            A job script that can be submitted to PBS.
        index
            The index of the job that is being run. This is used when
            specifying different resources for different jobs.

        Notes
        -----
        Currently, there is a problem that this will not properly cleanup.
        for example `ipengine ... &` will be detached and go on,
        normally a scheduler will take care of this.

        """
        job_script = textwrap.dedent(
            """\
            #!/bin/sh

            {extra_env_vars}

            {extra_script}

            {executor_specific}
            """,
        )

        return job_script.format(
            extra_env_vars=self.extra_env_vars(index=index),
            executor_specific=self._executor_specific("${NAME}", options, index=index),
            extra_script=self.extra_script(index=index),
            job_id_variable=self._JOB_ID_VARIABLE,
        )

    def queue(self, *, me_only: bool = True) -> dict[str, dict]:  # noqa: ARG002
        """Get the queue of the scheduler."""
        return self.mock_scheduler.queue()

    def start_job(self, name: str, *, index: int | None = None) -> None:
        """Start a job."""
        if index is not None:
            msg = "LocalMockScheduler does not support `index`."
            raise NotImplementedError(msg)
        name_prefix = name.rsplit("-", 1)[0]
        submit_cmd = f"{self.submit_cmd} {name} {self.batch_fname(name_prefix)}"
        run_submit(submit_cmd, name)

    def extra_scheduler(self, *, index: int | None = None) -> str:  # noqa: ARG002
        """Get the extra scheduler options."""
        msg = "extra_scheduler is not implemented."
        raise NotImplementedError(msg)



"""Test for the log parsing functions."""

from __future__ import annotations

import json
import os
import tempfile

from adaptive_scheduler.server_support import (
    DatabaseManager,
    _get_infos,
    parse_log_files,
)


def test_get_infos() -> None:
    """Test for _get_infos function."""
    # Prepare sample data
    log_data = [
        {
            "event": "current status",
            "timestamp": "2023-04-13 10:30.00",
            "elapsed_time": "0 days 01:30:00",
        },
        {
            "event": "current status",
            "timestamp": "2023-04-13 11:30.00",
            "elapsed_time": "0 days 02:30:00",
        },
    ]

    with tempfile.NamedTemporaryFile(mode="w+", delete=False) as f:
        for entry in log_data:
            f.write(json.dumps(entry) + "\n")
        f.flush()

        # Test only_last=True
        result = _get_infos(f.name, only_last=True)
        assert len(result) == 1
        assert result[0] == log_data[-1]

        # Test only_last=False
        result = _get_infos(f.name, only_last=False)
        assert len(result) == len(log_data)
        assert result == log_data[::-1]

    os.remove(f.name)  # noqa: PTH107


def test_parse_log_files(db_manager: DatabaseManager) -> None:
    """Test for parse_log_files function."""
    log_data = [
        {
            "event": "current status",
            "timestamp": "2023-04-13 10:30.00",
            "elapsed_time": "0 days 01:30:00",
        },
        {
            "event": "current status",
            "timestamp": "2023-04-13 11:30.00",
            "elapsed_time": "0 days 02:30:00",
        },
    ]

    with tempfile.NamedTemporaryFile(mode="w+", delete=False) as f:
        for entry in log_data:
            f.write(json.dumps(entry) + "\n")
        f.flush()

        # Prepare the database manager, scheduler, and job names
        job_name = "test_job"
        db_manager.start()
        # Add an entry in the database manager
        db_manager._choose_fname(job_name)
        db_manager._start_request("0", f.name, job_name)
        db_manager.scheduler.start_job(job_name)

        # Test parse_log_files
        df_result = parse_log_files(
            db_manager,
            db_manager.scheduler,
            only_last=False,
        )
        # Check that the returned DataFrame has the expected columns
        expected_columns = [
            "timestamp",
            "elapsed_time",
            "job_id",
            "log_fname",
            "fname",
            "is_pending",
            "is_done",
            "output_logs",
            "start_time",
            "state",
            "job_name",
        ]
        assert sorted(df_result.columns) == sorted(expected_columns)

        print(df_result)
        # Check that the returned DataFrame has the expected number of rows
        assert len(df_result) == len(log_data)

    os.remove(f.name)  # noqa: PTH107



from __future__ import annotations

import abc
import asyncio
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Coroutine


class ManagerAlreadyStartedError(Exception):
    """Raised when a manager is already started."""


class BaseManager(metaclass=abc.ABCMeta):
    def __init__(self) -> None:
        self.ioloop: asyncio.events.AbstractEventLoop | None = None
        self._coro: Coroutine | None = None
        self.task: asyncio.Task | None = None

    def start(self) -> BaseManager:
        if self.is_started:
            msg = f"{self.__class__} is already started!"
            raise ManagerAlreadyStartedError(msg)
        self._setup()
        self.ioloop = asyncio.get_event_loop()
        self._coro = self._manage()
        self.task = self.ioloop.create_task(self._coro)
        return self

    @property
    def is_started(self) -> bool:
        return self.task is not None

    def cancel(self) -> bool | None:
        if self.is_started:
            assert self.task is not None  # for mypy
            return self.task.cancel()
        return None

    def _setup(self) -> None:  # noqa: B027
        """Is run in the beginning of `self.start`."""

    @abc.abstractmethod
    async def _manage(self) -> None:  # pragma: no cover
        pass



"""Test `adaptive-scheduler-launcher`."""

from __future__ import annotations

import sys
from unittest.mock import MagicMock, patch

import pytest

# Import the functions you want to test from launcher.py
from adaptive_scheduler._server_support import launcher


# Test the _parse_args function
def test_parse_args() -> None:
    """Test that the _parse_args function returns the correct args."""
    test_args = [
        "launcher.py",
        "--n",
        "4",
        "--log-fname",
        "test.log",
        "--job-id",
        "123",
        "--name",
        "test",
        "--url",
        "http://example.com",
    ]

    # Store the original sys.argv
    original_argv = sys.argv

    # Replace sys.argv with the test arguments
    sys.argv = test_args

    # Call _parse_args and check the results
    args = launcher._parse_args()
    assert args.n == 4
    assert args.name == "test"

    # Restore the original sys.argv
    sys.argv = original_argv


# Test the _get_executor function
def test_get_executor() -> None:
    """Test that the _get_executor function returns the correct executor."""
    with patch("concurrent.futures.ProcessPoolExecutor") as mock_executor:
        mock_executor.return_value = MagicMock()
        executor = launcher._get_executor("process-pool", None, 4, None)
        assert executor is not None
        mock_executor.assert_called_with(max_workers=4)

    with pytest.raises(ValueError, match="Unknown executor_type"):
        launcher._get_executor("unknown", None, 4, "loky")  # type: ignore[arg-type]

    with patch("adaptive.runner.SequentialExecutor") as mock_executor:
        executor = launcher._get_executor("sequential", None, 4, None)
        assert executor is not None
        mock_executor.assert_called_once()
